/*
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifndef APACHE1_3 
/**
 * 
 * This is the hostinglimits module for apache 2.2.X
 * author Pookhov Sergey (pookhov@gmail.com)
 * author Igor Seletskiy <iseletsk@gmail.com>
 * How to make, install and setup see README file
 * 
 */
#include "apr_general.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"
#include <unistd.h>
#include "unixd.h"  

#include <lve/lve-ctl.h>
#include "stgitf.h"

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>

#define PREFIX "mod_hostinglimits:"

module AP_MODULE_DECLARE_DATA hostinglimits_module;
static apr_threadkey_t *key;
static apr_thread_once_t *once_object = NULL;
static apr_pool_t *ppool;

APR_DECLARE_OPTIONAL_FN(struct vh_info_t *, vh_info_lookup,
                        (conn_rec *, const char*));

static APR_OPTIONAL_FN_TYPE(vh_info_lookup) *pfn_vh_info_lookup = NULL;

// configuration data
typedef struct hostinglimits_module_cfg {
  unsigned int skip; 
  uint32_t lve_id;
  apr_array_header_t *allowed_handlers;
} hostinglimits_module_cfg;

static int lve_available = 0;
struct liblve *lve = NULL;
static int is_mutex_enabled = 0;



/**
 * This function returns directory config structure for current request
 */
static hostinglimits_module_cfg *hostinglimits_module_dconfig(const request_rec *r) {
  return (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config, 
							   &hostinglimits_module);
}


static void *hostinglimits_module_merge_config(apr_pool_t* p, void* BASE, void* ADD) {
  hostinglimits_module_cfg* base = BASE;
  hostinglimits_module_cfg* add = ADD;
  hostinglimits_module_cfg* cfg = (hostinglimits_module_cfg *) 
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  cfg->skip = (add->skip)?add->skip:base->skip;
  cfg->lve_id = (add->lve_id)?add->lve_id:base->lve_id;
  cfg->allowed_handlers = (add->allowed_handlers)?add->allowed_handlers:
    base->allowed_handlers;
  return cfg;
}

static void *hostinglimits_module_create_dir_config(apr_pool_t *p, 
						    char *dirspec) {
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) 
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  if(!cfg) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, NULL, 
		   PREFIX" not enough memory");      
    return NULL;
  }
  cfg->skip  = 0;
  cfg->lve_id = 0;
  cfg->allowed_handlers = NULL;
  return (void *) cfg;
}


static const char *set_lve_id(cmd_parms *cmd, void *mcfg,
			      const char *lve_id)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_id) {
    cfg->lve_id = (uint32_t)apr_atoi64(lve_id);;
  }
  return NULL;
}


static const char *set_handlers(cmd_parms *cmd, void *mcfg, const char *arg)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }
  if (!cfg->allowed_handlers) {
    cfg->allowed_handlers = apr_array_make(cmd->pool, 2, sizeof(char *));
  }
  *(const char **)apr_array_push(cfg->allowed_handlers) = arg;
  return NULL;
}
// describe used directives
static command_rec hostinglimits_module_directives[] = {
  AP_INIT_FLAG("SkipErrors", ap_set_flag_slot,  (void *) 
	       APR_OFFSETOF(hostinglimits_module_cfg, skip),
	       RSRC_CONF, "Allow apache to continue even if LVE is unavalable"),      
  AP_INIT_TAKE1("LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, "LVE Id"),
  AP_INIT_ITERATE("AllowedHandlers", set_handlers, NULL, RSRC_CONF, 
		  "A list of handlers which will be put in LVE"),

  {NULL}
};

static int init_post_config(apr_pool_t *pconf, apr_pool_t *plog,
			apr_pool_t *ptemp, server_rec *s) {
  pfn_vh_info_lookup = APR_RETRIEVE_OPTIONAL_FN(vh_info_lookup);
  if (lve_available) {
      ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, 
		   PREFIX"LVE mechanism enabled");      
  } else {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, 
		   PREFIX"LVE mechanism disabled, LVE is unavailable");
  }
  
  return OK;
}

static int match_handler(apr_array_header_t *allowed_handlers, const char *handler) {
  int num_names;
  char **names_ptr;
  int accept = 0;
  if (allowed_handlers) {
      names_ptr = (char **)allowed_handlers->elts;
      num_names = allowed_handlers->nelts;
    } else {
      return 0;
    }
    // match handle
    for (; num_names; ++names_ptr, --num_names) {
      if (!strcmp(handler, *names_ptr)) {
	accept=1;
	break;
      }
    }
    return accept;
}

/**
 * Where are main functionality in this function.
 * This funcion performs after reading request headers
 * and before other proceccing.
 * 
 * 
 */
static int enter_lve_handler(request_rec *r) {
  // get directory config info
  uint32_t *p_cookie = NULL;
  apr_status_t st;
  hostinglimits_module_cfg *cfg = hostinglimits_module_dconfig(r);
  if (!lve_available || !cfg) {
    return DECLINED;
  }
  /* decline if cookie is set ( p_cookie != 0) */
  st=apr_threadkey_private_get((void **)&p_cookie, key);
  if(st != APR_SUCCESS || (p_cookie!=NULL && *p_cookie != 0)) {
     return DECLINED;
  }
  //lets find handlers for which HostingLimits should run
  if (!match_handler(cfg->allowed_handlers, r->handler)) {
    return DECLINED;
  }

  //enter into lve
  uid_t    uid = -1;
  int lve_id = cfg->lve_id;

  struct vh_info_t* vh = NULL;
  if(pfn_vh_info_lookup) {
    vh = pfn_vh_info_lookup(r->connection, r->hostname);
  }
  //is h2e available?
  if (vh) {
    uid = (uid_t)vh->m_user_id;
  } else {
    ap_unix_identity_t * ugid = ap_run_get_suexec_identity(r);
    if (ugid == NULL) {
      uid = -1;
    } else {
      uid = ugid->uid;
    }
    if (lve_id < 1) {
      lve_id = uid;
    }
  }

  if (lve_id > 0) {
    p_cookie = (uint32_t*)apr_pcalloc(r->pool, sizeof(uint32_t));
    if(!p_cookie) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
		   PREFIX"Can't allocate memory");
      return HTTP_SERVICE_UNAVAILABLE;
    }
    int rc = lve_enter(lve, lve_id, -1, -1, p_cookie); //todo add limits
    if(rc) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
		   PREFIX"Error on LVE enter %d", rc);
      return HTTP_SERVICE_UNAVAILABLE;
    }
    st = apr_threadkey_private_set(p_cookie, key);
    
    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
		   PREFIX"Unable to set LVE cookie %d", st);
    }
  } 

  return DECLINED;
}   


static int leave_lve_handler(request_rec *r) {
  uint32_t *p_cookie;
  apr_status_t st;
  if (!lve_available) {
    return DECLINED;
  }

  st=apr_threadkey_private_get((void **)&p_cookie, key);
  /* if cookie is not initialized or if it is not 0 -- don't try to leave */
  if (st != APR_SUCCESS || p_cookie == NULL || *p_cookie == 0){
    return DECLINED;
  }

  hostinglimits_module_cfg  *cfg = hostinglimits_module_dconfig(r);
	 
  int rc = lve_leave(lve, p_cookie);
  if(!rc) {
    /* reset cookie to 0 */
    *p_cookie = 0;
    st = apr_threadkey_private_set(NULL, key);
    
    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
		   PREFIX"Unable to reset LVE cookie %d", st);
    }
    return DECLINED;
  } 
  ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, 
	       PREFIX"Error exiting %d", rc);
  //error exiting, lets kill it
  exit(-1);
}

static apr_status_t destroy_key_pool(void * dummy)
{
  apr_threadkey_private_delete(key);
  apr_pool_destroy(ppool);
  ppool = 0;
  return APR_SUCCESS;
}
        

static void hostinglimits_child_init(apr_pool_t *p, server_rec *s){
  if (APR_SUCCESS == apr_pool_create ((&ppool), 0)){
   apr_pool_cleanup_register (p, 0, destroy_key_pool, destroy_key_pool);
   apr_threadkey_private_create(&key, NULL, ppool);
  } else {
    ap_log_error(APLOG_MARK, APLOG_ERR, OK, s, 
	       PREFIX"Can't create Key pool");
    exit(-1);
  }

}

// function for hook register
static void hostinglimits_module_register_hooks(apr_pool_t *p) {
  int rc;
  if (lve == NULL) {
    lve = apr_palloc(p, lve_instance_init(NULL));
    rc = lve_instance_init(lve);
    if (rc != 0) {
      lve_available = 0;
    } else {
      lve_available = 1;
      // register clean up LVE
      apr_pool_cleanup_register(p, lve, (void*)lve_instance_destroy, apr_pool_cleanup_null);
    }
  }

  static const char * const aszPre[] = { "mod_include.c", "mod_php.c", "mod_cgi.c", NULL };
	
  ap_hook_post_config(init_post_config, NULL, NULL, APR_HOOK_MIDDLE);
  // hook executes before request prforming
  ap_hook_handler(enter_lve_handler, NULL, aszPre, APR_HOOK_REALLY_FIRST);
  // hook executes after data generation and logging
  ap_hook_log_transaction(leave_lve_handler, NULL, NULL, APR_HOOK_LAST);
  ap_hook_child_init(hostinglimits_child_init,NULL,NULL,APR_HOOK_MIDDLE);

}

/**
 * Describing structure of Apache module
 */
module AP_MODULE_DECLARE_DATA hostinglimits_module = {
  STANDARD20_MODULE_STUFF,
  hostinglimits_module_create_dir_config,   /* create directory config */
  hostinglimits_module_merge_config,        /* merging directory config */
  NULL,                                     /* create server config */
  NULL,                                     /* merging server config */
  hostinglimits_module_directives,          /* mapping configuration directives */
  hostinglimits_module_register_hooks       /* registering hooks */
};

#else

#define CORE_PRIVATE

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>

#include <lve/lve-ctl.h>
#include "stgitf.h" 

#define PREFIX "mod_hostinglimits:"

module MODULE_VAR_EXPORT hostinglimits_module; 

//static APR_OPTIONAL_FN_TYPE(vh_info_lookup) *pfn_vh_info_lookup = NULL; 

static int lve_available = 0;
struct liblve *lve = NULL;
static uint32_t p_cookie = 0;
                                                                                                                                                     
// configuration data
typedef struct hostinglimits_module_cfg {                                                                                                                    
  unsigned int skip;
  uint32_t lve_id;
  array_header *allowed_handlers;     
} hostinglimits_module_cfg;

static hostinglimits_module_cfg *hostinglimits_module_dconfig(request_rec *r){                                                                                                                                                            
  return (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config,&hostinglimits_module);
}

static void *hostinglimits_module_create_dir_config(pool *p, char *dirspec)
{

  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *)ap_pcalloc(p, sizeof(hostinglimits_module_cfg));
  if(!cfg)  {
      ap_log_error(APLOG_MARK, APLOG_ERR, NULL, PREFIX" not enough memory");
      return NULL;
  }                                                                                                                                                          
  cfg->skip  = 0;
  cfg->lve_id = 0;
  cfg->allowed_handlers = NULL;    
  return (void *) cfg; 
    
}

static void *hostinglimits_module_merge_config(pool* p, void* BASE, void* ADD) {                                                                       

  hostinglimits_module_cfg* base = BASE;
  hostinglimits_module_cfg* add = ADD;
  hostinglimits_module_cfg* cfg = (hostinglimits_module_cfg *)ap_pcalloc(p, sizeof(hostinglimits_module_cfg));
  cfg->skip = (add->skip)?add->skip:base->skip;
  cfg->lve_id = (add->lve_id)?add->lve_id:base->lve_id;
  cfg->allowed_handlers = (add->allowed_handlers)?add->allowed_handlers:base->allowed_handlers;
  return (void *)cfg;
                                                                                                                                                
}

static const char *set_lve_id(cmd_parms *cmd, void *mcfg, const char *lve_id)
{                                                                                                                                                            
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_id) {
      cfg->lve_id = (uint32_t)atoi(lve_id);
  }                                                                                                                                                          
  return NULL;
}  

static const char *set_skip(cmd_parms *cmd, void *mcfg, const char *skip){                                                                                                                                                            
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (skip) {
      cfg->skip = (unsigned int)atoi(skip);
  }                                                                                                                                                          
  return NULL;
}  

static const char *set_handlers(cmd_parms *cmd, void *mcfg, const char *arg){                                                                                                                                                            
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }                                                                                                                                                          
  if (!cfg->allowed_handlers) {
    cfg->allowed_handlers = ap_make_array(cmd->pool, 2, sizeof(char *));
  }                                                                                                                                                          
  *(const char **)ap_push_array(cfg->allowed_handlers) = arg;
  return NULL;
}

static int hostinglimits_module_exit(request_rec *r)
{     
   if (!p_cookie) {
       return DECLINED;
   }                                                                                                                                                          
   if (lve_available) {
     hostinglimits_module_cfg  *cfg = hostinglimits_module_dconfig(r);
                                                
    if(!cfg) {
     return DECLINED;
    }                                                                                                                                                        
    int rc=lve_leave(lve, &p_cookie);
    if(!rc) {
     p_cookie = 0;
     return DECLINED;
    }                                                                                                                   
    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,PREFIX"Error exiting %d", rc);
    //error exiting, lets kill it                                                                                                                            
    exit(-1);
  }                                                                                                                                                          
                                                                                                                                                                                                                                                                                                                                                                                                                
  return DECLINED;
}

static void hostinglimits_initializer(server_rec *s, pool *p){                                                                                               
  int rc;
  if (lve == NULL) {
     lve = ap_palloc(p, lve_instance_init(NULL));
     rc = lve_instance_init(lve);
     if (rc != 0) {
       lve_available = 0;
     } else {
       lve_available = 1;
       ap_register_cleanup(p, lve, (void*)lve_instance_destroy, ap_null_cleanup);
     }                                                                                                                                                        
  }
  //pfn_vh_info_lookup = APR_RETRIEVE_OPTIONAL_FN(vh_info_lookup);  
  //i can't find APR_RETRIEVE_OPTIONAL_FN analog for apache 1.3. 
  //it comes from apr_utils, will be apr_utils correct work with apache 1.3?
  //i disabled this function
  if (lve_available) {
    ap_log_error(APLOG_MARK, APLOG_NOTICE, s, "LVE mechanism enabled");
    //printf("LVE mechanism enabled\n");
  } else {
    ap_log_error(APLOG_MARK, APLOG_ERR, s, "LVE mechanism disabled, LVE is unavailable");
    //printf("LVE mechanism disabled, LVE is unavailable\n");
  }
} 

static int match_handler(array_header *allowed_handlers, const char *handler) {
  int num_names;
  char **names_ptr;
  int accept = 0;
  if (allowed_handlers && handler) {
    names_ptr = (char **)allowed_handlers->elts;
    num_names = allowed_handlers->nelts;
  } else {
    return 0;
  }
  // match handle
  for (; num_names; ++names_ptr, --num_names) {
    if (!strcmp(handler, *names_ptr)) {
        accept=1;
        break;
    }
  }
  return accept;
}

static int enter_lve_fixups(request_rec *r) {
  // get directory config info        

  hostinglimits_module_cfg *cfg = hostinglimits_module_dconfig(r);
  if(p_cookie || !cfg) {
    // we are already in lve, or config not initiated - request declined
    return DECLINED;   
  }
  if (!lve_available && !cfg->skip) {
    ap_log_error(APLOG_MARK, APLOG_ERR, r->server, PREFIX"LVE is unavailable LVE");
    return HTTP_SERVICE_UNAVAILABLE;
  }
  if (lve_available) {
    //lets find handlers for which HostingLimits should run
    if ((!match_handler(cfg->allowed_handlers, r->handler))&&(!match_handler(cfg->allowed_handlers, r->content_type))) {
        return DECLINED;
    }

    //enter into lve
    uid_t    uid = -1;
    int lve_id = cfg->lve_id;

    uid = r->server->server_uid;
    if (lve_id < 1) {
        lve_id = uid;
    }
    if (lve_id > 0) {
      int rc = lve_enter(lve, lve_id, -1, -1, &p_cookie); //todo add limits
      if(rc) {
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, PREFIX"Error on LVE enter %d", rc);
        p_cookie = 0;
        return HTTP_SERVICE_UNAVAILABLE;
      }
    } 
  }
  return DECLINED;
}

static command_rec hostinglimits_module_directives[] = {
  {"SkipErrors", set_skip, NULL, RSRC_CONF, TAKE1, "Allow apache to continue even if LVE is unavalable"}, 
  {"LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, TAKE1, "LVE Id"},
  {"AllowedHandlers", set_handlers, NULL, RSRC_CONF, ITERATE,"A list of handlers which will be put in LVE"},
  {NULL}                                         
};


module hostinglimits_module = {
  STANDARD_MODULE_STUFF,
  hostinglimits_initializer,                     /* initializer */
  hostinglimits_module_create_dir_config,  /* dir config creator */
  hostinglimits_module_merge_config,       /* dir merger --- default is to override */
  NULL,   /* server config */
  NULL,  /* merge server config */
  hostinglimits_module_directives,                 /* command table */
  NULL,             /* handlers */
  NULL,    /* filename translation */
  NULL,                     /* check_user_id */
  NULL,                     /* check auth */
  NULL,                     /* check access */
  NULL,         /* type_checker */
  enter_lve_fixups,                     /* fixups */
  hostinglimits_module_exit,        /* logger */
  NULL,                      /* header parser */
  NULL,      /* [3] header parser */                                                                                                      
  NULL,         /* process initializer */                                                                                                    
  NULL,         /* process exit/cleanup */                                                                                                                                                                                                                            
  NULL   /* [1] post read_request handling */                                                                                          
};                                             

#undef CORE_PRIVATE


#endif
